'''
test cases for keys functionality
'''

from clodss import clodss
import os

db = clodss.StrictRedis(
    os.path.realpath(os.path.dirname(__file__) + '/../data'),
    decode_responses=True
)

hkey = 'some-map'

def test_hget():
    key = 'key_get'
    assert db.hget(hkey, key) is None

def test_hset():
    key = 'key_set'
    value = 'some value'
    db.hset(hkey, key, value)
    assert db.hget(hkey, key) == value


def test_hset_existing():
    key = 'key_set_existing'
    value = 'some other value'
    db.hset(hkey, key, 'random value')
    db.hset(hkey, key, value)
    assert db.hget(hkey, key) == value

def test_hdel_existing():
    key = 'key_set_existing'
    db.hset(hkey, key, 'random value')
    assert db.hdel(hkey, key) == 1
    assert db.hget(hkey, key) is None


def test_hdel_nonexisting_hash():
    assert db.hdel('nonexisting_map', 1) == 0


def test_hdel_nonexisting_key():
    key = 'key_set_existing'
    db.hset(hkey, key, 'random value')
    assert db.hdel(hkey, 'nonexisting_key') == 0


def test_hkeys_nonexisting():
    assert db.hkeys('nonexisting_map') == []


def test_hvalues_nonexisting():
    assert db.hvalues('nonexisting_map') == []


def test_hkeys_hvalues():
    db.delete(hkey)
    db.hset(hkey, 1, 'v1')
    db.hset(hkey, 2, 'v2')
    db.hset(hkey, 3, 'v3')
    assert set(db.hkeys(hkey)) == {'1', '2', '3'}
    assert set(db.hvalues(hkey)) == {'v1', 'v2', 'v3'}


def test_hgetall():
    db.delete(hkey)
    db.hset(hkey, 1, 'v1')
    db.hset(hkey, 2, 'v2')
    db.hset(hkey, 3, 'v3')
    assert db.hgetall(hkey) == {'1': 'v1', '2': 'v2', '3': 'v3'}


def test_hgetall_nonexisting():
    assert db.hgetall('nonexisting') == {}


def test_hmset():
    db.delete(hkey)
    db.hmset(hkey, {'1': 'v1', '2': 'v2', '3': 'v3'})
    assert db.hget(hkey, 1) == 'v1'
    assert db.hget(hkey, 2) == 'v2'
    assert db.hget(hkey, 3) == 'v3'


def test_hmget():
    db.delete(hkey)
    db.hmset(hkey, {'1': 'v1', '2': 'v2', '3': 'v3'})
    assert db.hmget(hkey, *range(5)) == [None, 'v1', 'v2', 'v3', None]



'''
CLODSS: Close* Data-Structures Store

clodss is a data-structures on-disk store with an API largly compatible with
redis. The goal is to develop a store with the simplicity of the redis API
which scales beyond memory capacity, allows harnessing multi-core processors,
and does not burden accesses with network latency.

Homepage: https://github.com/codomatech/clodss
Author: Codoma.tech Advanced Technologies
        https://www.codoma.tech/
License: BSD

'''

from .clodss import StrictRedis



# -*- coding: utf-8 -*-

'''
clodss: hashmap data-structure
'''

from .common import SEP

def _augkey(hkey, key):
    return f'{hkey}{SEP}h{SEP}{key}'


def _db(instance, hkey):
    return instance.router.connection(hkey).db()


def hset(instance, hkey, key, val):
    'https://redis.io/commands/hset'
    db = _db(instance, hkey)
    db[_augkey(hkey, key)] = val
    return 1


def hget(instance, hkey, key):
    'https://redis.io/commands/hget'
    db = _db(instance, hkey)
    try:
        return instance.makevalue(db[_augkey(hkey, key)])
    except KeyError:
        return None


def hdel(instance, hkey, key):
    'https://redis.io/commands/hdel'
    db = _db(instance, hkey)
    key = _augkey(hkey, key)
    if not key in db:
        return 0
    del db[key]
    return 1


def hkeys(instance, hkey):
    'https://redis.io/commands/hkeys'
    db = _db(instance, hkey)
    keys = []
    prefix = _augkey(hkey, '').encode('utf-8')
    for k, _ in db[prefix:]:
        if not k.startswith(prefix):
            break
        keys.append(instance.makevalue(k[len(prefix):]))
    return keys

def hvalues(instance, hkey):
    'https://redis.io/commands/hkeys'
    db = _db(instance, hkey)
    vals = []
    prefix = _augkey(hkey, '').encode('utf-8')
    for k, v in db[prefix:]:
        if not k.startswith(prefix):
            break
        vals.append(instance.makevalue(v))
    return vals

def hgetall(instance, hkey):
    'https://redis.io/commands/hgetall'
    db = _db(instance, hkey)
    items = {}
    prefix = _augkey(hkey, '').encode('utf-8')
    for k, v in db[prefix:]:
        if not k.startswith(prefix):
            break
        items[instance.makevalue(k[len(prefix):])] = instance.makevalue(v)
    return items


def hmset(instance, hkey, mapping):
    'https://redis.io/commands/hmset'
    for k, v in mapping.items():
        hset(instance, hkey, k, v)


def hmget(instance, hkey, *keys):
    'https://redis.io/commands/hmget'
    return [hget(instance, hkey, k) for k in keys]



'''
test cases for keys functionality
'''

import os
import time

import pytest
from clodss import clodss
from clodss.common import SEP

db = clodss.StrictRedis(
    os.path.realpath(os.path.dirname(__file__) + '/../data'),
    decode_responses=True
)


def test_invalid_key():
    with pytest.raises(TypeError):
        db.get()
    with pytest.raises(ValueError):
        db.get(f'a{SEP}b')


def test_change_type():
    key = 'byteskey'
    db.set(key, 123)
    with pytest.raises(ValueError):
        db.hset(key, 1, 2)
    db.knownkeys = {}
    with pytest.raises(ValueError):
        db.hset(key, 1, 2)


def test_util_funcs():
    assert type(db.stats()) is dict
    assert type(db.router.poolstatus()) is dict


def test_get():
    key = 'key_get'
    assert db.get(key) is None

def test_set():
    key = 'key_set'
    value = 'some value'
    db.set(key, value)
    assert db.get(key) == value


def test_set_existing():
    key = 'key_set_existing'
    value = 'some other value'
    db.set(key, value)
    assert db.get(key) == value


def test_del():
    key = 'key_del'
    value = 'some other value'
    db.set(key, value)
    db.delete(key)
    assert db.get(key) is None


def test_incr_inited():
    key = 'key_incr'
    db.set(key, 5)
    db.incr(key)
    assert db.get(key) == '6'


def test_incr_uninited():
    key = 'key_incr_uninited'
    db.delete(key)
    db.incr(key)
    assert db.get(key) == '1'


def test_incr_nonint():
    key = 'key_incr_nonint'
    db.set(key, 'ab')
    with pytest.raises(TypeError):
        db.incr(key)


def test_incrby_inited():
    key = 'key_incrby'
    db.set(key, 5)
    db.incrby(key, 20)
    assert db.get(key) == '25'


def test_incrby_uninited():
    key = 'key_incrby_uninited'
    db.delete(key)
    db.incrby(key, 33)
    assert db.get(key) == '33'


def test_incrby_nonint():
    key = 'key_incrby_nonint'
    db.set(key, 'ab')
    with pytest.raises(TypeError):
        db.incrby(key, 44)


def test_decr_inited():
    key = 'key_decr'
    db.set(key, 5)
    db.decr(key)
    assert db.get(key) == '4'


def test_decr_uninited():
    key = 'key_decr_uninited'
    db.delete(key)
    db.decr(key)
    assert db.get(key) == '-1'


def test_decr_nonint():
    key = 'key_decr_nonint'
    db.set(key, 'ab')
    with pytest.raises(TypeError):
        db.decr(key)


def test_decrby_inited():
    key = 'key_decrby'
    db.set(key, 50)
    db.decrby(key, 20)
    assert db.get(key) == '30'


def test_decrby_uninited():
    key = 'key_decrby_uninited'
    db.delete(key)
    db.decrby(key, 33)
    assert db.get(key) == '-33'


def test_decrby_nonint():
    key = 'key_decrby_nonint'
    db.set(key, 'ab')
    with pytest.raises(TypeError):
        db.decrby(key, 44)


def test_expire_nonexitent():
    assert db.expire('non-existing', 5) == 0


def test_expire():
    key = 'key_expire'
    value = 'some value'
    db.set(key, value)
    assert db.expire(key, .5) == 1
    time.sleep(.1)
    assert db.get(key) == value
    time.sleep(.5)
    assert db.get(key) is None


def test_persist_nonexitent():
    assert db.persist('non-existing') == 0


def test_persist_nonscheduled():
    key = 'key_persist_nonscheduled'
    db.set(key, 123)
    assert db.persist(key) == 0


def test_persist():
    key = 'key_persist'
    value = 'some value'
    db.set(key, value)
    db.expire(key, .5)
    time.sleep(.1)
    assert db.persist(key) == 1
    time.sleep(.45)
    assert db.get(key) == value


def test_flushdb():
    for i in range(10):
        db.set(f'bytes-{i}', i)
        db.hset(f'map-{i}', 'key', i)
        db.lpush(f'list-{i}', i)
    db.flushdb()
    assert len(db.keys()) == 0


def test_keys():
    db.flushdb()
    keys = set()
    for i in range(10):
        db.set(f'bytes-{i}', i)
        db.hset(f'map-{i}', 'key', i)
        db.lpush(f'list-{i}', i)
        keys |= {f'bytes-{i}', f'map-{i}', f'list-{i}'}
    assert set(db.keys()) == keys
    assert set(db.keys('map-*')) == {k for k in keys if k.startswith('map-')}


def test_scan():
    db.flushdb()
    keys = set()
    for i in range(10):
        db.set(f'bytes-{i}', i)
        db.hset(f'map-{i}', 'key', i)
        db.lpush(f'list-{i}', i)
        keys |= {f'bytes-{i}', f'map-{i}', f'list-{i}'}
    cur, res = db.scan()
    assert set(res) == keys
    assert cur == True
    assert db.scan(cur) == []
    _, res = db.scan(match='map-*')
    assert set(res) == {k for k in keys if k.startswith('map-')}
    #assert set(db.keys('map-*')) == {k for k in keys if k.startswith('map-')}



# -*- coding: utf-8 -*-

'''
clodss is a data-structures on-disk store with an API largly compatible with
redis. The goal is to develop a store with the simplicity of the redis API
which scales beyond memory capacity, allows harnessing multi-core processors,
and does not burden accesses with network latency.
'''

import logging
import time
import os
import __main__
from .router import Router
from . import hashmaps
from . import lists
from . import keys
from .common import SEP


def wrapmethod(method, stats=None):
    '''
    - guards all clodss methods with a key-scoped lock
    - performs sanity checks on key
    - enures the key has not expired
    '''
    def wrapper(*args, **kwargs):
        globalmethod = method.__name__ in ('keys', 'scan', 'flushdb')

        instance = args[0]
        if globalmethod:
            key = '﹁'
        else:
            if len(args) < 2:
                raise TypeError('too few parameters, `key` is required')
            key = args[1]

            if SEP in key:
                raise ValueError(f'`key` contains invalid character(s): {SEP}')
        dtype = instance.keydtype(key)
        methodtype = method.__name__[0].encode('utf-8')
        if method.__name__ in ('rpush', 'rpop'):
            methodtype = b'l'
        elif methodtype not in (b'l', b'h'):
            methodtype = ''
        #print(method.__name__, key, dtype, methodtype, globalmethod)
        if method.__name__ != 'delete' and dtype is not None \
            and not globalmethod and dtype != methodtype:
            raise ValueError('incompatible operation `%s` on %s' %(
                method.__name__, dtype))

        if stats is not None:
            t1 = time.perf_counter()
        if not globalmethod:
            instance.checkexpired(key, enforce=True)
        result = method(*args, **kwargs)
        if stats is not None:
            t = time.perf_counter() - t1
            avg, n = stats.get(method.__name__, (0, 0))
            avg = (avg * n + t) / (n + 1)
            stats[method.__name__] = (avg, n + 1)
        return result
    wrapper.__name__ = method.__name__
    return wrapper


class StrictRedis:
    'main clodss class'
    def __init__(
            self, dbpath: str = None, db: int = 0, spread_factor: int = 2,
            decode_responses: bool = False, benchmark: bool = True) -> None:
        if dbpath is None:
            try:
                fname = __main__.__file__
            except AttributeError:
                fname = '<unknown>'
            d = os.path.dirname(fname)
            if d == '':
                d = '.'
            dbpath = os.path.realpath(f'{d}/./clodss-data')
        logging.info('clodss db path: %s', dbpath)
        self.decode = decode_responses
        dbpath = os.path.join(dbpath, '%02d' % db)
        os.makedirs(dbpath, exist_ok=True)
        self._dbpath = dbpath
        self.router = Router(dbpath, spread_factor)
        self.knownkeys = {}
        self.keystoexpire = {}
        self._stats = {} if benchmark else None

        modules = [keys, lists, hashmaps]
        for module in modules:
            for attr in dir(module):
                if attr.startswith('_'):
                    continue
                method = getattr(module, attr)
                if not callable(method):
                    continue
                if attr == 'sēt':
                    # `set` is a reserved keyword
                    attr = 'set'
                setattr(StrictRedis, attr, wrapmethod(method, self._stats))

    def stats(self):
        'gets benchmarking statistics'
        return self._stats

    def makevalue(self, v):
        'makes value respecting decode_responses setting'
        if self.decode:
            return v.decode('utf-8')
        return v

    def dbpath(self):
        'gets base path for database files'
        return self._dbpath

    def keydtype(self, key):
        'get key data type'
        db = self.router.connection(key).db()
        closestkey = b''
        for k, _ in db[key:]:
            closestkey = k
            break
        if closestkey != key.encode('utf-8') and not closestkey.startswith(
                f'{key}{SEP}'.encode('utf-8')):
            return None
        if not SEP.encode('utf-8') in closestkey:
            return ''
        return closestkey.split(SEP.encode('utf-8'))[1]

    def checkexpired(self, key, enforce=False):
        'checks if a key expired and removes it if so'
        db = self.router.connection(key).db()
        try:
            exp = self.keystoexpire.get(key)
            if not exp:
                try:
                    exp = db[f'{SEP}expire{SEP}{key}']
                except KeyError:
                    pass
            if exp:
                t = float(exp)
                now = time.time()
                #print('now', now)
                if now > t:
                    print('expiring', key, type(key))
                    if not enforce:
                        return True
                    del db[f'{SEP}expire{SEP}{key}'.encode('utf-8')]
                    for k, _ in db[key:]:
                        print('checking', k, key, k == key)
                        if k == key.encode('utf-8') or k.startswith(
                                f'{key}{SEP}'.encode('utf-8')):
                            print('deleting', k, f'{SEP}expire{SEP}{k}')
                            del db[k]
                        else:
                            break
                    if key in self.knownkeys:
                        del self.knownkeys[key]

                    return True
                return 'scheduled'
            return False
        finally:
            pass

    def reset(self):
        'clears a database and all cached information'
        self.router.reset()
        self.knownkeys = {}
        self.keystoexpire = {}



# -*- coding: utf-8 -*-

'''
clodss: keys-related functions
'''

import re
import time

from .common import SEP, _clearexpired


def _keyexists(db, key):
    for k, _ in db[key.encode('utf-8'):]:
        if k == key.encode('utf-8') or k.startswith(
                f'{key}{SEP}'.encode('utf-8')):
            return True
        break
    return False


def get(instance, key):
    'https://redis.io/commands/get'
    db = instance.router.connection(key).db()
    try:
        return instance.makevalue(db[key])
    except KeyError:
        return None


def sēt(instance, key, value):
    'https://redis.io/commands/set'
    db = instance.router.connection(key).db()
    db[key] = value


def delete(instance, key):
    'https://redis.io/commands/del'
    db = instance.router.connection(key).db()
    for k, _ in db[key:]:
        if k == key.encode('utf-8') or k.startswith(
                f'{key}{SEP}'.encode('utf-8')):
            del db[k]
        else:
            break


def expire(instance, key, duration: float) -> int:
    'https://redis.io/commands/expire'
    db = instance.router.connection(key).db()
    if not _keyexists(db, key):
        print('key doesnt exist', key)
        return 0
    expiretime = duration + time.time()
    db[f'{SEP}expire{SEP}{key}'] = expiretime
    instance.keystoexpire[key] = expiretime
    return 1


def persist(instance, key) -> int:
    'https://redis.io/commands/persist'
    db = instance.router.connection(key).db()
    if not _keyexists(db, key):
        return 0
    if instance.checkexpired(key) != 'scheduled':
        return 0
    _clearexpired(instance, db, key)
    return 1


def incr(instance, key, amount=1):
    'https://redis.io/commands/incr'
    db = instance.router.connection(key).db()
    try:
        val = db[key]
    except KeyError:
        val = 0
    try:
        val = int(val)
    except ValueError:
        raise TypeError('invalid numeric %s' %val)
    newval = val + amount
    db[key] = newval
    return newval


def incrby(instance, key, amount):
    'https://redis.io/commands/incrby'
    return incr(instance, key, amount)


def decr(instance, key, amount=1):
    'https://redis.io/commands/decr'
    db = instance.router.connection(key).db()
    try:
        val = db[key]
    except KeyError:
        val = 0
    try:
        val = int(val)
    except ValueError:
        raise TypeError('invalid numeric %s' %val)
    newval = int(val) - amount
    db[key] = newval
    return newval


def decrby(instance, key, amount):
    'https://redis.io/commands/decrby'
    return decr(instance, key, amount)


def flushdb(instance):
    'https://redis.io/commands/flushdb'
    instance.reset()


def keys(instance, pattern='*'):
    '''
    https://redis.io/commands/keys
    NOTE: there is no guarantee the keys returned are complete or precise, i.e.
    some existing keys might be missing and some retured keys might be deleted.
    For now we favor performance for this specific command, please use hkeys as
    a reliable alternative.
    '''
    allkeys = []
    pattern = pattern.replace('*', '.*')
    regex = re.compile(pattern)
    for db in instance.router.allconnections():
        for k, _ in db.db():
            k = k.split(SEP.encode('utf-8'))[0]
            #print('adding key', k)
            if regex.match(k.decode('utf-8')):
                allkeys.append(instance.makevalue(k))
    return allkeys


def scan(instance, cursor=None, match='*'):
    '''
    https://redis.io/commands/scan
    currently is the same as `keys`
    '''

    if cursor not in (None, True):
        raise ValueError('invalid cursor %r' %cursor)

    if cursor is True:
        return []

    return True, keys(instance, match)



# -*- coding: utf-8 -*-

'''
common utilities
'''


SEP = chr(0x2c3)

def _clearexpired(instance, db, key):
    if instance.checkexpired(key) in (True, 'scheduled'):
        del db[f'{SEP}expire{SEP}{key}']
        if key in instance.keystoexpire:
            del instance.keystoexpire[key]


class ProblematicSymbols:
    'remove symbols that could cause sql injection or use reserved words'

    FWD = {
        'sqlite_': 'sqlite﹁',
        '`': "﹁'﹁",
    }

    BWD = {v: k for k, v in FWD.items()}

    @staticmethod
    def remove(text):
        'remove symbols'
        for kw, rep in ProblematicSymbols.FWD.items():
            text = text.replace(kw, rep)
        return text

    @staticmethod
    def restore(text):
        'resotre symbols'
        for kw, rep in ProblematicSymbols.BWD.items():
            text = text.replace(kw, rep)
        return text



# -*- coding: utf-8 -*-

'''
router.py: provides the Router class to map keys to a corresponding database.
routing is configured by a spread factor
'''

import os
import hashlib
import uuid
import lsm

class DBConnection:
    'a database connection used within a pool'
    def __init__(self, fname):
        self._id = uuid.uuid1().hex
        self.free = True
        self._db = lsm.LSM(fname)

    def db(self):
        'db object'
        return self._db

    def id(self):
        'connection id'
        return self._id


class Router:
    'main routing class, maps keys to a db based on a spread factor'
    EXT = 'clodssdb'

    def __init__(self, dbpath, factor=2, poolsize=3):
        '''
        dbpath: where to store the data files
        factor: partitioning factor, the higher it is, the more spread your
        data will be. this improves concurrency but also increases the number
        of open files. defaults to 2
        poolsize: pool size per data file, defaults to 3
        '''
        self.factor = factor
        self.dbpath = dbpath
        self.poolsize = poolsize
        self.connections = {}

    def _alldbs(self):
        return sorted([
            os.path.join(self.dbpath, f)
            for f in os.listdir(self.dbpath)
            if f.endswith(Router.EXT)
        ])

    def reset(self):
        'clears the database and closes all connections'
        for db in self._alldbs():
            os.unlink(db)
        self.connections = {}

    def connection(self, key: bytes):
        'gets a new connection'
        db = hashlib.sha1(key.encode('utf-8')).hexdigest()[:self.factor]
        return DBConnection(os.path.join(self.dbpath, f'{db}.{Router.EXT}'))

    def allconnections(self, offset: int = 0):
        '''
        generator of connections to all available dbs
        offset: start offset in the alphabetically sorted list of dbs
        '''
        dbs = self._alldbs()[offset:]
        for db in dbs:
            yield DBConnection(db)

    def poolstatus(self):
        'gets pooling status'
        return {k: [conn.free for conn in conns]
                for k, conns in self.connections.items()}



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="clodss",
    version="0.0.5",
    author="Codoma.tech Team",
    author_email="info@codoma.tech",
    description="On-Disk data-structures store with redis-like API.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/codomatech/clodss",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'lsm-db',
    ],
)



# -*- coding: utf-8 -*-

'''
clodss: list data-structure
'''

from .common import SEP

MAX_DIGITS = 12


MIDDLE_INDEX = int(10 ** (MAX_DIGITS / 2))


def _augkey(lkey):
    return f'{lkey}{SEP}l{SEP}'


def _maxkey(lkey, db, asint=True):
    prefix = _augkey(lkey).encode('utf-8')
    maxkey = None
    for k, _ in db[prefix::True]:
        maxkey = k
        break
    if maxkey is None or not maxkey.startswith(prefix):
        return None
    return maxkey if not asint else int(maxkey[len(prefix):])


def _minkey(lkey, db, asint=True):
    prefix = _augkey(lkey).encode('utf-8')
    minkey = None
    for k, _ in db[prefix:]:
        minkey = k
        break
    if minkey is None or not minkey.startswith(prefix):
        return None
    return minkey if not asint else int(minkey[len(prefix):])


def llen(instance, key) -> int:
    'https://redis.io/commands/llen'
    db = instance.router.connection(key).db()
    prefix = _augkey(key).encode('utf-8')
    n = 0
    for k, _ in db[prefix:]:
        if not k.startswith(prefix):
            break
        n += 1
    return n


def rpush(instance, key, val):
    'https://redis.io/commands/rpush'
    db = instance.router.connection(key).db()
    prefix = _augkey(key).encode('utf-8')
    maxkey = _maxkey(key, db)
    index = MIDDLE_INDEX if maxkey is None else maxkey
    indexstr = f'%0{MAX_DIGITS}d' % (index + 1)
    if len(indexstr) > MAX_DIGITS:
        # TO DO: cleanup to use deleted indices
        raise Exception('list is too big')
    db[prefix + indexstr.encode('utf-8')] = val


def lpush(instance, key, val):
    'https://redis.io/commands/lpush'
    db = instance.router.connection(key).db()
    prefix = _augkey(key).encode('utf-8')
    minkey = _minkey(key, db)
    index = MIDDLE_INDEX if minkey is None else minkey
    index = index - 1
    if index < 0:
        # TO DO: cleanup to use deleted indices
        raise Exception('list is too big')
    indexstr = f'%0{MAX_DIGITS}d' % (index)
    db[prefix + indexstr.encode('utf-8')] = val


def rpop(instance, key):
    'https://redis.io/commands/rpop'
    db = instance.router.connection(key).db()
    maxkey = _maxkey(key, db, False)
    if maxkey is None:
        return None
    v = db[maxkey]
    del db[maxkey]
    return instance.makevalue(v)


def lpop(instance, key):
    'https://redis.io/commands/lpop'
    db = instance.router.connection(key).db()
    minkey = _minkey(key, db, False)
    if minkey is None:
        return None
    v = db[minkey]
    del db[minkey]
    return instance.makevalue(v)


def _normalizeindex(index, lkey, instance):
    if index >= 0:
        return index
    l = llen(instance, lkey)
    if l == 0:
        return 0
    return (index + (-index // l + 1) * l) % l


def _lindex(lkey, index: int, instance):
    db = instance.router.connection(lkey).db()
    prefix = _augkey(lkey).encode('utf-8')
    index = _normalizeindex(index, lkey, instance)
    i = 0
    key = None
    for k, _ in db[prefix:]:
        if not k.startswith(prefix):
            break
        if i == index:
            key = k
            break
        i += 1
    return key


def lindex(instance, key, index: int):
    'https://redis.io/commands/lindex'
    db = instance.router.connection(key).db()
    k = _lindex(key, index, instance)
    if k is None:
        return None
    return instance.makevalue(db[k])


def lset(instance, key, index: int, value: bytes):
    'https://redis.io/commands/lset'
    db = instance.router.connection(key).db()
    k = _lindex(key, index, instance)
    if k is None:
        return
    db[k] = value


def lrange(instance, key, start: int, end: int):
    'https://redis.io/commands/lrange'
    db = instance.router.connection(key).db()
    k0 = _lindex(key, start, instance)
    k1 = _lindex(key, end, instance)
    if k0 is None:
        return None
    if k1 is None:
        k1 = _maxkey(key, db, False)
    return [instance.makevalue(v) for _, v in db[k0:k1]]


def ltrim(instance, key, start: int, end: int) -> None:
    'https://redis.io/commands/ltrim'
    db = instance.router.connection(key).db()
    prefix = _augkey(key).encode('utf-8')
    start = _normalizeindex(start, key, instance)
    end = _normalizeindex(end, key, instance)
    if start > end:
        instance.delete(key)
        return
    i = 0
    for k, _ in db[prefix:]:
        if not k.startswith(prefix):
            break
        if not start <= i <= end:
            del db[k]
        i += 1


def lrem(instance, key, count: int, value: bytes) -> None:
    'https://redis.io/commands/lrem'
    db = instance.router.connection(key).db()
    prefix = _augkey(key).encode('utf-8')
    todel = []
    iterable = db[prefix::True] if count < 0 else db[prefix:]
    limit = int(10**MAX_DIGITS) if count == 0 else abs(count)
    for k, v in iterable:
        if v == value.encode('utf-8'):
            todel.append(k)
        if len(todel) == limit:
            break
        if not k.startswith(prefix):
            break
    for k in todel:
        del db[k]
    return len(todel)


def linsert(instance, key, where, refvalue, value) -> int:
    'https://redis.io/commands/linsert'
    db = instance.router.connection(key).db()
    if instance.keydtype(key) is None:
        return 0
    prefix = _augkey(key).encode('utf-8')
    marker = SEP * 10
    n = 0
    found = False
    cache = {}
    for k, v in db[prefix:]:
        if not k.startswith(prefix):
            break
        n += 1
        if found:
            if len(cache) >= 42:
                for ck, cv in cache.items():
                    db[ck] = cv
                cache = {}
            cache[k] = prev
            prev = v
        else:
            if v == refvalue.encode('utf-8'):
                found = True
                instance.rpush(key, marker)
                if where == 'before':
                    prev = refvalue
                    cache[k] = value
                else:
                    prev = value
    if found:
        for ck, cv in cache.items():
            db[ck] = cv
        lastkey = _maxkey(key, db, False)
        print('last key=', lastkey)
        db[lastkey] = prev
    return -1 if not found else n
